/* vim:set et sts=4: */
/* bus - The Input Bus
 * Copyright (C) 2008-2010 Peng Huang <shawn.p.huang@gmail.com>
 * Copyright (C) 2008-2010 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <stdlib.h>
#include "ibusenginedesc.h"
#include "ibusxml.h"

enum {
    LAST_SIGNAL,
};

enum {
    PROP_0 = 0,
    PROP_NAME,
    PROP_LONGNAME,
    PROP_DESCRIPTION,
    PROP_LANGUAGE,
    PROP_LICENSE,
    PROP_AUTHOR,
    PROP_ICON,
    PROP_LAYOUT,
    PROP_HOTKEYS,
    PROP_RANK,
};


/* IBusEngineDescPriv */
struct _IBusEngineDescPrivate {
    gpointer    pad;
    gchar      *hotkeys;
};
typedef struct _IBusEngineDescPrivate IBusEngineDescPrivate;

#define IBUS_ENGINE_DESC_GET_PRIVATE(o)  \
   (G_TYPE_INSTANCE_GET_PRIVATE ((o), IBUS_TYPE_ENGINE_DESC, IBusEngineDescPrivate))

// static guint            _signals[LAST_SIGNAL] = { 0 };

/* functions prototype */
static void         ibus_engine_desc_set_property   (IBusEngineDesc         *desc,
                                                     guint                   prop_id,
                                                     const GValue           *value,
                                                     GParamSpec             *pspec);
static void         ibus_engine_desc_get_property   (IBusEngineDesc         *desc,
                                                     guint                   prop_id,
                                                     GValue                 *value,
                                                     GParamSpec             *pspec);
static void         ibus_engine_desc_destroy        (IBusEngineDesc         *desc);
static gboolean     ibus_engine_desc_serialize      (IBusEngineDesc         *desc,
                                                     IBusMessageIter        *iter);
static gboolean     ibus_engine_desc_deserialize    (IBusEngineDesc         *desc,
                                                     IBusMessageIter        *iter);
static gboolean     ibus_engine_desc_copy           (IBusEngineDesc         *dest,
                                                     const IBusEngineDesc   *src);
static gboolean     ibus_engine_desc_parse_xml_node (IBusEngineDesc         *desc,
                                                     XMLNode                *node);

G_DEFINE_TYPE (IBusEngineDesc, ibus_engine_desc, IBUS_TYPE_SERIALIZABLE)


static void
ibus_engine_desc_class_init (IBusEngineDescClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    IBusObjectClass *object_class = IBUS_OBJECT_CLASS (klass);
    IBusSerializableClass *serializable_class = IBUS_SERIALIZABLE_CLASS (klass);

    gobject_class->set_property = (GObjectSetPropertyFunc) ibus_engine_desc_set_property;
    gobject_class->get_property = (GObjectGetPropertyFunc) ibus_engine_desc_get_property;
    object_class->destroy = (IBusObjectDestroyFunc) ibus_engine_desc_destroy;

    serializable_class->serialize   = (IBusSerializableSerializeFunc) ibus_engine_desc_serialize;
    serializable_class->deserialize = (IBusSerializableDeserializeFunc) ibus_engine_desc_deserialize;
    serializable_class->copy        = (IBusSerializableCopyFunc) ibus_engine_desc_copy;

    g_string_append (serializable_class->signature, "ssssssssuusu");

    g_type_class_add_private (klass, sizeof (IBusEngineDescPrivate));

    /* install properties */
    /**
     * IBusEngineDesc:name:
     *
     * The name of engine description
     */
    g_object_class_install_property (gobject_class,
                    PROP_NAME,
                    g_param_spec_string ("name",
                        "description name",
                        "The name of engine description",
                        NULL,
                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    /**
     * IBusEngineDesc:longname:
     *
     * The longname of engine description
     */
    g_object_class_install_property (gobject_class,
                    PROP_LONGNAME,
                    g_param_spec_string ("longname",
                        "description longname",
                        "The longname of engine description",
                        NULL,
                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    /**
     * IBusEngineDesc:description:
     *
     * The description of engine description
     */
    g_object_class_install_property (gobject_class,
                    PROP_DESCRIPTION,
                    g_param_spec_string ("description",
                        "description description",
                        "The description of engine description",
                        NULL,
                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    /**
     * IBusEngineDesc:language:
     *
     * The language of engine description
     */
    g_object_class_install_property (gobject_class,
                    PROP_LANGUAGE,
                    g_param_spec_string ("language",
                        "description language",
                        "The language of engine description",
                        NULL,
                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    /**
     * IBusEngineDesc:license:
     *
     * The license of engine description
     */
    g_object_class_install_property (gobject_class,
                    PROP_LICENSE,
                    g_param_spec_string ("license",
                        "description license",
                        "The license of engine description",
                        NULL,
                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    /**
     * IBusEngineDesc:author:
     *
     * The author of engine description
     */
    g_object_class_install_property (gobject_class,
                    PROP_AUTHOR,
                    g_param_spec_string ("author",
                        "description author",
                        "The author of engine description",
                        NULL,
                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    /**
     * IBusEngineDesc:icon:
     *
     * The icon of engine description
     */
    g_object_class_install_property (gobject_class,
                    PROP_ICON,
                    g_param_spec_string ("icon",
                        "description icon",
                        "The icon of engine description",
                        NULL,
                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    /**
     * IBusEngineDesc:layout:
     *
     * The layout of engine description
     */
    g_object_class_install_property (gobject_class,
                    PROP_LAYOUT,
                    g_param_spec_string ("layout",
                        "description layout",
                        "The layout of engine description",
                        NULL,
                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    /**
     * IBusEngineDesc:hotkeys:
     *
     * The hotkeys of engine description
     */
    g_object_class_install_property (gobject_class,
                    PROP_HOTKEYS,
                    g_param_spec_string ("hotkeys",
                        "description hotkeys",
                        "The hotkeys of engine description",
                        NULL,
                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    /**
     * IBusEngineDesc:rank:
     *
     * The rank of engine description
     */
    g_object_class_install_property (gobject_class,
                    PROP_RANK,
                    g_param_spec_uint ("rank",
                        "description rank",
                        "The lank of eingine description",
                        0,
                        G_MAXUINT,
                        0,
                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
}

static void
ibus_engine_desc_init (IBusEngineDesc *desc)
{
    IBusEngineDescPrivate *priv;

    desc->name = NULL;
    desc->longname = NULL;
    desc->description = NULL;
    desc->language = NULL;
    desc->license = NULL;
    desc->author = NULL;
    desc->icon = NULL;
    desc->layout = NULL;
    desc->rank = 0;

    priv = IBUS_ENGINE_DESC_GET_PRIVATE(desc);
    priv->hotkeys = NULL;
}

static void
ibus_engine_desc_destroy (IBusEngineDesc *desc)
{
    IBusEngineDescPrivate *priv;

    g_free (desc->name);
    g_free (desc->longname);
    g_free (desc->description);
    g_free (desc->language);
    g_free (desc->license);
    g_free (desc->author);
    g_free (desc->icon);
    g_free (desc->layout);

    priv = IBUS_ENGINE_DESC_GET_PRIVATE(desc);
    g_free (priv->hotkeys);

    IBUS_OBJECT_CLASS (ibus_engine_desc_parent_class)->destroy (IBUS_OBJECT (desc));
}

static void
ibus_engine_desc_set_property (IBusEngineDesc *desc,
                               guint           prop_id,
                               const GValue   *value,
                               GParamSpec     *pspec)
{
    IBusEngineDescPrivate *priv;
    priv = IBUS_ENGINE_DESC_GET_PRIVATE (desc);

    switch (prop_id) {
    case PROP_NAME:
        g_assert (desc->name == NULL);
        desc->name = g_strdup (g_value_get_string (value));
        break;
    case PROP_LONGNAME:
        g_assert (desc->longname == NULL);
        desc->longname = g_strdup (g_value_get_string (value));
        break;
    case PROP_DESCRIPTION:
        g_assert (desc->description == NULL);
        desc->description  = g_strdup (g_value_get_string (value));
        break;
    case PROP_LANGUAGE:
        g_assert (desc->language == NULL);
        desc->language = g_strdup (g_value_get_string (value));
        break;
    case PROP_LICENSE:
        g_assert (desc->license == NULL);
        desc->license = g_strdup (g_value_get_string (value));
        break;
    case PROP_AUTHOR:
        g_assert (desc->author == NULL);
        desc->author = g_strdup (g_value_get_string (value));
        break;
    case PROP_ICON:
        g_assert (desc->icon == NULL);
        desc->icon = g_strdup (g_value_get_string (value));
        break;
    case PROP_LAYOUT:
        g_assert (desc->layout == NULL);
        desc->layout = g_strdup (g_value_get_string (value));
        break;
    case PROP_RANK:
        g_assert (desc->rank == 0);
        desc->rank = g_value_get_uint (value);
        break;
    case PROP_HOTKEYS:
        g_assert (priv->hotkeys == NULL);
        priv->hotkeys = g_strdup (g_value_get_string (value));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (desc, prop_id, pspec);
    }
}

const gchar *
ibus_engine_desc_get_hotkeys (IBusEngineDesc *desc)
{
    IBusEngineDescPrivate *priv;
    priv = IBUS_ENGINE_DESC_GET_PRIVATE (desc);

    return priv->hotkeys;
}

static void
ibus_engine_desc_get_property (IBusEngineDesc *desc,
                               guint           prop_id,
                               GValue         *value,
                               GParamSpec     *pspec)
{
    switch (prop_id) {
    case PROP_NAME:
        g_value_set_string (value, desc->name);
        break;
    case PROP_LONGNAME:
        g_value_set_string (value, desc->longname);
        break;
    case PROP_DESCRIPTION:
        g_value_set_string (value, desc->description);
        break;
    case PROP_LANGUAGE:
        g_value_set_string (value, desc->language);
        break;
    case PROP_LICENSE:
        g_value_set_string (value, desc->license);
        break;
    case PROP_AUTHOR:
        g_value_set_string (value, desc->author);
        break;
    case PROP_ICON:
        g_value_set_string (value, desc->icon);
        break;
    case PROP_LAYOUT:
        g_value_set_string (value, desc->layout);
        break;
    case PROP_RANK:
        g_value_set_uint (value, desc->rank);
        break;
    case PROP_HOTKEYS:
        g_value_set_string (value, ibus_engine_desc_get_hotkeys (desc));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (desc, prop_id, pspec);
    }
}

static gboolean
ibus_engine_desc_serialize (IBusEngineDesc  *desc,
                            IBusMessageIter *iter)
{
    gboolean retval;
    guint prop_id;
    IBusEngineDescPrivate *priv;

    retval = IBUS_SERIALIZABLE_CLASS (ibus_engine_desc_parent_class)->serialize ((IBusSerializable *)desc, iter);
    g_return_val_if_fail (retval, FALSE);

    retval = ibus_message_iter_append (iter, G_TYPE_STRING, &desc->name);
    g_return_val_if_fail (retval, FALSE);

    retval = ibus_message_iter_append (iter, G_TYPE_STRING, &desc->longname);
    g_return_val_if_fail (retval, FALSE);

    retval = ibus_message_iter_append (iter, G_TYPE_STRING, &desc->description);
    g_return_val_if_fail (retval, FALSE);

    retval = ibus_message_iter_append (iter, G_TYPE_STRING, &desc->language);
    g_return_val_if_fail (retval, FALSE);

    retval = ibus_message_iter_append (iter, G_TYPE_STRING, &desc->license);
    g_return_val_if_fail (retval, FALSE);

    retval = ibus_message_iter_append (iter, G_TYPE_STRING, &desc->author);
    g_return_val_if_fail (retval, FALSE);

    retval = ibus_message_iter_append (iter, G_TYPE_STRING, &desc->icon);
    g_return_val_if_fail (retval, FALSE);

    retval = ibus_message_iter_append (iter, G_TYPE_STRING, &desc->layout);
    g_return_val_if_fail (retval, FALSE);

    retval = ibus_message_iter_append (iter, G_TYPE_UINT, &desc->rank);
    g_return_val_if_fail (retval, FALSE);

    priv = IBUS_ENGINE_DESC_GET_PRIVATE(desc);

    prop_id = PROP_HOTKEYS;
    ibus_message_iter_append (iter, G_TYPE_UINT, &prop_id);
    retval = ibus_message_iter_append (iter, G_TYPE_STRING, &priv->hotkeys);
    g_return_val_if_fail (retval, FALSE);

    prop_id = 0;
    ibus_message_iter_append (iter, G_TYPE_UINT, &prop_id);

    return TRUE;
}

static gboolean
ibus_engine_desc_deserialize (IBusEngineDesc  *desc,
                              IBusMessageIter *iter)
{
    gboolean retval;
    IBusEngineDescPrivate *priv;
    guint prop_id;
    gchar *str;

    retval = IBUS_SERIALIZABLE_CLASS (ibus_engine_desc_parent_class)->deserialize ((IBusSerializable *)desc, iter);
    g_return_val_if_fail (retval, FALSE);

    retval = ibus_message_iter_get (iter, G_TYPE_STRING, &str);
    g_return_val_if_fail (retval, FALSE);
    ibus_message_iter_next (iter);
    desc->name = g_strdup (str);

    retval = ibus_message_iter_get (iter, G_TYPE_STRING, &str);
    g_return_val_if_fail (retval, FALSE);
    ibus_message_iter_next (iter);
    desc->longname = g_strdup (str);

    retval = ibus_message_iter_get (iter, G_TYPE_STRING, &str);
    g_return_val_if_fail (retval, FALSE);
    ibus_message_iter_next (iter);
    desc->description = g_strdup (str);

    retval = ibus_message_iter_get (iter, G_TYPE_STRING, &str);
    g_return_val_if_fail (retval, FALSE);
    ibus_message_iter_next (iter);
    desc->language = g_strdup (str);

    retval = ibus_message_iter_get (iter, G_TYPE_STRING, &str);
    g_return_val_if_fail (retval, FALSE);
    ibus_message_iter_next (iter);
    desc->license = g_strdup (str);

    retval = ibus_message_iter_get (iter, G_TYPE_STRING, &str);
    g_return_val_if_fail (retval, FALSE);
    ibus_message_iter_next (iter);
    desc->author = g_strdup (str);

    retval = ibus_message_iter_get (iter, G_TYPE_STRING, &str);
    g_return_val_if_fail (retval, FALSE);
    ibus_message_iter_next (iter);
    desc->icon = g_strdup (str);

    retval = ibus_message_iter_get (iter, G_TYPE_STRING, &str);
    g_return_val_if_fail (retval, FALSE);
    ibus_message_iter_next (iter);
    desc->layout = g_strdup (str);

    retval = ibus_message_iter_get (iter, G_TYPE_UINT, &desc->rank);
    g_return_val_if_fail (retval, FALSE);
    ibus_message_iter_next (iter);

    priv = IBUS_ENGINE_DESC_GET_PRIVATE(desc);

    while (1) {
        if (ibus_message_iter_get (iter, G_TYPE_UINT, &prop_id) == FALSE) {
            break;
        }
        if (prop_id == 0) {
            break;
        }
        switch (prop_id) {
        case PROP_HOTKEYS:
            retval = ibus_message_iter_get (iter, G_TYPE_STRING, &str);
            if (retval == FALSE) {
                break;
            }
            priv->hotkeys = g_strdup (str);
            break;
        default:
            retval = FALSE;
        }
        g_return_val_if_fail (retval, FALSE);
        ibus_message_iter_next (iter);
    }

    return TRUE;
}

static gboolean
ibus_engine_desc_copy (IBusEngineDesc       *dest,
                       const IBusEngineDesc *src)
{
    gboolean retval;
    IBusEngineDescPrivate *priv_dest;
    IBusEngineDescPrivate *priv_src;

    retval = IBUS_SERIALIZABLE_CLASS (ibus_engine_desc_parent_class)->copy ((IBusSerializable *)dest,
                                 (IBusSerializable *)src);
    g_return_val_if_fail (retval, FALSE);


    dest->name          = src->name ? g_strdup (src->name) : NULL;
    dest->longname      = src->longname ? g_strdup (src->longname) : NULL;
    dest->description   = src->description ? g_strdup (src->description) : NULL;
    dest->language      = src->language ? g_strdup (src->language) : NULL;
    dest->license       = src->license ? g_strdup (src->license) : NULL;
    dest->author        = src->author ? g_strdup (src->author) : NULL;
    dest->icon          = src->icon ? g_strdup (src->icon) : NULL;
    dest->layout        = src->layout ? g_strdup (src->layout) : NULL;

    priv_dest = IBUS_ENGINE_DESC_GET_PRIVATE(dest);
    priv_src = IBUS_ENGINE_DESC_GET_PRIVATE(src);

    priv_dest->hotkeys  = priv_src->hotkeys ? g_strdup (priv_src->hotkeys) : NULL;
    return TRUE;
}

#define g_string_append_indent(string, indent)  \
    {                                           \
        gint i;                                 \
        for (i = 0; i < (indent); i++) {        \
            g_string_append (string, "    ");   \
        }                                       \
    }

void
ibus_engine_desc_output (IBusEngineDesc *desc,
                         GString        *output,
                         gint            indent)
{
    IBusEngineDescPrivate *priv;

    priv = IBUS_ENGINE_DESC_GET_PRIVATE(desc);

    g_string_append_indent (output, indent);
    g_string_append (output, "<engine>\n");
#define OUTPUT_ENTRY(field, element, desc)                                              \
    {                                                                                   \
        gchar *escape_text = g_markup_escape_text (desc->field ? desc->field : "", -1); \
        g_string_append_indent (output, indent + 1);                                    \
        g_string_append_printf (output, "<"element">%s</"element">\n",                  \
                                escape_text);                                           \
        g_free (escape_text);                                                           \
    }
#define OUTPUT_ENTRY_1(name) OUTPUT_ENTRY(name, #name, desc)
#define OUTPUT_ENTRY_2(name) OUTPUT_ENTRY(name, #name, priv)
    OUTPUT_ENTRY_1(name);
    OUTPUT_ENTRY_1(longname);
    OUTPUT_ENTRY_1(description);
    OUTPUT_ENTRY_1(language);
    OUTPUT_ENTRY_1(license);
    OUTPUT_ENTRY_1(author);
    OUTPUT_ENTRY_1(icon);
    OUTPUT_ENTRY_1(layout);
    OUTPUT_ENTRY_2(hotkeys);
    g_string_append_indent (output, indent + 1);
    g_string_append_printf (output, "<rank>%u</rank>\n", desc->rank);
#undef OUTPUT_ENTRY
#undef OUTPUT_ENTRY_1
#undef OUTPUT_ENTRY_2
    g_string_append_indent (output, indent);
    g_string_append (output, "</engine>\n");
}

static gboolean
ibus_engine_desc_parse_xml_node (IBusEngineDesc *desc,
                                XMLNode       *node)
{
    GList *p;
    IBusEngineDescPrivate *priv;

    priv = IBUS_ENGINE_DESC_GET_PRIVATE(desc);
    for (p = node->sub_nodes; p != NULL; p = p->next) {
        XMLNode *sub_node = (XMLNode *) p->data;

#define PARSE_ENTRY(field_name, element_name, desc)             \
        if (g_strcmp0 (sub_node->name, element_name) == 0) {    \
            g_free (desc->field_name);                          \
            desc->field_name = g_strdup (sub_node->text);       \
            continue;                                           \
        }
#define PARSE_ENTRY_1(name) PARSE_ENTRY(name, #name, desc)
#define PARSE_ENTRY_2(name) PARSE_ENTRY(name, #name, priv)
        PARSE_ENTRY_1(name);
        PARSE_ENTRY_1(longname);
        PARSE_ENTRY_1(description);
        PARSE_ENTRY_1(language);
        PARSE_ENTRY_1(license);
        PARSE_ENTRY_1(author);
        PARSE_ENTRY_1(icon);
        PARSE_ENTRY_1(layout);
        PARSE_ENTRY_2(hotkeys);
#undef PARSE_ENTRY
#undef PARSE_ENTRY_1
#undef PARSE_ENTRY_2
        if (g_strcmp0 (sub_node->name , "rank") == 0) {
            desc->rank = atoi (sub_node->text);
            continue;
        }
        g_warning ("<engines> element contains invalidate element <%s>", sub_node->name);
    }
    return TRUE;
}

IBusEngineDesc *
ibus_engine_desc_new (const gchar *name,
                      const gchar *longname,
                      const gchar *description,
                      const gchar *language,
                      const gchar *license,
                      const gchar *author,
                      const gchar *icon,
                      const gchar *layout)
{
    return ibus_engine_desc_new2 ("name", name,
                                  "longname", longname,
                                  "description", description,
                                  "language", language,
                                  "license", license,
                                  "author", author,
                                  "icon", icon,
                                  "layout", layout,
                                  NULL);
}

IBusEngineDesc *
ibus_engine_desc_new2 (const gchar *first_property_name, ...)
{
    va_list var_args;
    IBusEngineDesc *desc;
    IBusEngineDescPrivate *priv;

    g_assert (first_property_name);

    va_start (var_args, first_property_name);
    desc = (IBusEngineDesc *) g_object_new_valist (IBUS_TYPE_ENGINE_DESC,
                                                   first_property_name,
                                                   var_args);
    va_end (var_args);

    g_assert (desc->name);

    if (desc->longname == NULL) {
        desc->longname = g_strdup ("");
    }
    if (desc->description == NULL) {
        desc->description   = g_strdup ("");
    }
    if (desc->language == NULL) {
        desc->language      = g_strdup ("dummy");
    }
    if (desc->license == NULL) {
        desc->license       = g_strdup ("");
    }
    if (desc->author == NULL) {
        desc->author        = g_strdup ("");
    }
    if (desc->icon == NULL) {
        desc->icon          = g_strdup ("ibus-engine");
    }
    if (desc->layout == NULL) {
        desc->layout        = g_strdup ("us");
    }

    priv = IBUS_ENGINE_DESC_GET_PRIVATE(desc);
    if (priv->hotkeys == NULL) {
        priv->hotkeys       = g_strdup ("");
    }

    return desc;
}

IBusEngineDesc *
ibus_engine_desc_new_from_xml_node (XMLNode      *node)
{
    g_assert (node);

    IBusEngineDesc *desc;

    if (G_UNLIKELY (g_strcmp0 (node->name, "engine") != 0)) {
        return NULL;
    }

    desc = (IBusEngineDesc *)g_object_new (IBUS_TYPE_ENGINE_DESC, NULL);

    if (!ibus_engine_desc_parse_xml_node (desc, node)) {
        g_object_unref (desc);
        desc = NULL;
    }

    return desc;
}
